#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['_ParamBase', 'Param']

import sys
import types
import logging

import pyutilib.misc
from component import Component
from set_types import Any
from indexed_component import IndexedComponent
from misc import apply_indexed_rule, apply_parameterized_indexed_rule
from numvalue import value
from plugin import ParamRepresentationFactory
from pyutilib.component.core import alias
from sets import _BaseSet
from param_repn import _ParamData

logger = logging.getLogger('coopr.pyomo')


class _ParamBase(IndexedComponent):
    """A parameter value, which may be defined over a index"""

    """ Constructor
        Arguments:
           name       The name of this parameter
           index      The index set that defines the distinct parameters.
                         By default, this is None, indicating that there
                         is a single parameter.
           within     A set that defines the type of values that
                         each parameter must be.
           validate   A rule for validating this parameter w.r.t. data
                         that exists in the model
           default    A set that defines default values for this parameter
           rule       A rule for setting up this parameter with existing model
                         data
           nochecking If true, various checks regarding valid domain and index
                      values are skipped. Only intended for use by algorithm
                      developers - not modelers!
    """
    def __init__(self, *args, **kwd):
        self._initialize = kwd.pop('initialize', None )
        self._initialize = kwd.pop('rule', self._initialize )
        self._validate   = kwd.pop('validate', None )
        self.domain      = kwd.pop('within', Any )
        self.nochecking  = kwd.pop('nochecking',False)
        defaultval       = kwd.pop('default', None )
        self.repn_type = kwd.pop('repn', 'pyomo_dict')
        #
        kwd.setdefault('ctype', Param)
        IndexedComponent.__init__(self, *args, **kwd)
        #
        self._default = _ParamData( self.name, self.domain, value(defaultval))
        #
        self.clear()

    def clear(self):
        """Clear the data in this component"""
        self._data = ParamRepresentationFactory( self.repn_type, args=(self,) )
        if self._data is None:
            raise ValueError, "Unknown parameter representation '%s'" % self.repn_type

    def initialize(self, data):
        self._initialize = data

    def pprint(self, ostream=None, verbose=False):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, "  ",self.name,":",
        if not self.doc is None:
            print >>ostream, self.doc
            print >>ostream, "  ",
        print >>ostream, "\tSize="+str(len(self)),
        print >>ostream, "\tDomain="+self.domain.name
        if None in self.keys():
            if self._constructed is True:
                print >>ostream, "\t", self._data(None)
            else:
                print >>ostream, "\t", "Not constructed"
        else:
            tmp=self._data.keys(nondefault=True)
            tmp.sort()
            for key in tmp:
                val = self._data(key)
                print >>ostream, "\t"+str(key)+" : "+str(val)
        if self._default.value is not None and not self._default.value is None:
            print >>ostream, "\tdefault: "+str(self._default.value)

    #
    # a utiliy to extract all index-value pairs defining this
    # parameter, returned as a dictionary. useful in many
    # contexts, in which key iteration and repeated __getitem__
    # calls are too expensive to extract the contents of a parameter.
    # NOTE: We are presently not careful if an index has not been
    #       explicitly defined - issues around defaults are bound
    #       to crop up.
    #
    def extract_values(self):
        return self._data.extract_values()

    # a utility to set values of a parameter "in bulk", using a dictionary of index<->value pairs.
    # far more efficient in some contexts than the approach of doing things by-index in an outer loop.    
    def store_values(self, new_values):
        return self._data.store_values(new_values)

    def set_default(self, value):
        self._default.value = value
        self._data.set_default(value)

    def as_numeric(self):
        if None in self._data:
            return self._data[None]
        return self

    def is_expression(self):
        return False

    def is_relational(self):
        return False

    def __setitem__(self, ndx, val):
        # Convert a tuple index value ...
        _type = type(ndx)
        if _type is tuple or _type is list:
            if self._ndim == 1:
                # An index for a single-dimensional parameter is just the first tuple value
                ndx = ndx[0]
            else:
                # Flatten index for a multi-dimensional non-singleton parameter
                ndx = tuple(pyutilib.misc.flatten(ndx))

        # a somewhat gory way to see if you have a singleton - the len()=1 check is needed to avoid "None" 
        # being passed illegally into set membership validation rules. the nochecking screen saves 
        # significant time here, as the len() computation is remarkably expensive.
        if (self.nochecking is False) and (len(self._index) == 1) and (None in self._index) and (ndx is not None):  
            # allow for None indexing only if this is truly a singleton
            msg = "Cannot set an array value in the simple parameter '%s'"
            raise KeyError, msg % self.name
        #
        if (self.nochecking is False) and (ndx not in self._index):
            msg = "Cannot set the value of array parameter '%s' with invalid index '%s'"
            raise KeyError, msg % ( self.name, str(ndx) )
        #
        self._data[ndx] = val
        #
        # This is at the end, so the user can write their validation
        # tests as if the data was already added.
        #
        if (self.nochecking is False) and (not self._valid_indexed_value(val, ndx, False)):
            msg = "Invalid parameter value: %s[%s] = '%s', value type=%s"
            raise ValueError, msg % ( self.name, str(ndx), str(val), str(type(val)) )

    def construct(self, data=None):
        """ Apply the rule to construct values in this set """
        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing Param, name="+self.name+", from data="+`data`)
        if self._constructed:
            raise IOError, "Cannot reconstruct parameter '%s'" % self.name
            return
        self.clear()
        self._constructed=True
        #
        # Code optimization with local variables
        #
        _name=self.name
        _domain=self.domain
        #
        self._data.update_index()
        rule = getattr(self,'rule',None)
        if not rule is None:
            self._initialize=rule
        #
        # Construct using the initial data or the data loaded from an
        # external source.  Note:  data values will be queried in to following
        # order:
        #   1) the 'data' dictionary
        #   2) the self._initialize dictionary
        #   3) the default value
        #
        if data is not None or type(self._initialize) is not types.FunctionType:
            #
            # Singleton (non-indexed) parameter
            #
            #print self._data, self._initialize, self._default.value, self._index
            if type(self._index) is dict:
                error = False
                if data is not None and None in data.keys():
                    self.value=data[None]
                elif not self._initialize is None:
                    if type(self._initialize) is not dict:
                        self.value=self._initialize
                    elif None in self._initialize:
                        self.value=self._initialize[None]
                    else:
                        error = True
                elif self._default.value is not None:
                    self.value=self._default.value
                else:
                    error = True
                if error:
                    msg = "Attempting to construct parameter '%s' without "   \
                          'parameter data'
                    raise ValueError, msg % _name

                #self.pprint()
                #print self.value#, self[None].value
                self[None] = self.value
                self._valid_indexed_value(self.value,(),True)
            else:
                #
                # Set external data values (if provided)
                #
                if data is not None:
                    #print 'z'
                    for key in data:
                        if type(key) is tuple and len(key)==1:
                            tmpkey=key[0]
                        else:
                            tmpkey=key
                        self.__setitem__(key, data[key])
                #
                # Initialize with initialization data.
                #
                elif self._initialize is not None:
                    #print 'y'
                    if type(self._initialize) is dict:
                        for key in self._initialize:
                            self.__setitem__(key, self._initialize[key])
                    else:
                        self._data.set_item(val=self._initialize)
        #
        # Construct using the rule
        #
        elif type(self._initialize) is types.FunctionType:
            if (type(self._index) is dict) and (None in self._index):
                # singleton
                self.value = self._initialize(self._model())
                self[None] = self.value
            else:
                # non-singleton
                for index in self._index:
                    # IMPORTANT: Do *not* call self.__setitem__ here - we already know the index is
                    #            valid (and flattened), and validating indices is far too expensive to 
                    #            waste time doing it again. however, we still need to validate the
                    #            value generated by the rule, relative to the parameter domain.
                    param_value = apply_indexed_rule(self, self._initialize, self._model(), index)
                    if not ((self.domain is Any) or (param_value in self.domain)):
                        msg = "Invalid value '%s' specified via rule initializer for parameter '%s', index '%s'"
                        raise ValueError, msg % (str(param_value), self.name, index)
                    self._data[index] = param_value
        #
        if self.repn_type != 'pyomo_dict':
            self._model().__dict__[self.name] = self._data.model_repn()

    def _valid_indexed_value(self, value, index, use_exception):
        if index is ():
            if None not in self._index:
                msg = "Invalid index '%s' for parameter '%s'"
                raise ValueError, msg % ( str(None), self.name )
        elif type(index) is tuple and len(index)==1:
            if index not in self._index and index[0] not in self._index:
                msg = "Invalid index '%s' for parameter '%s'"
                raise ValueError, msg % ( str(index[0]), self.name)
        elif index not in self._index:
            msg = "Invalid index '%s' for parameter '%s'"
            raise ValueError, msg % ( str(index), self.name)
        #
        if self._validate is not None:
            if index is None:
                tmp = ()
            elif index == ():
                tmp = (None,)*self._ndim 
            else:
                tmp = index

            if apply_parameterized_indexed_rule(
                self, self._validate, self._model(), value, tmp ):
                return True
        #
        elif self.domain is Any or value in self.domain:
            return True
        #
        if use_exception:           #pragma:nocover
            msg = "Invalid value '%s' for parameter '%s'"
            raise ValueError, msg % ( str(value), self.name )
        #
        return False

    def data(self):
        return self._data.data()


class _ParamElement(_ParamBase, _ParamData):

    def __init__(self, *args, **kwds):

        repn_name = kwds.get('repn', 'pyomo_dict')

        _ParamBase.__init__(self, *args, **kwds)
        _ParamData.__init__(self, kwds.get('name',None), kwds.get('within',Any), kwds.get('default',None))

        if repn_name == 'pyomo_dict':
            self._data.set_item(None, self)
        else:
            self._data.set_item(None, self._default.value)

    def __getstate__(self):
        parambase_result = _ParamBase.__getstate__(self)
        paramvalue_result = _ParamData.__getstate__(self)
        return dict(parambase_result.items() + paramvalue_result.items())   

    def __setstate__(self, state):
        for (slot_name, value) in state.iteritems():
            setattr(self, slot_name, value)
        
    def __len__(self):
        return 1

    def keys(self):
        return [None]

    def __getitem__(self, key):
        if not key is None:
            raise KeyError, "Undefined key %s for parameter %s" % (key, self.name)
        return _ParamBase.__getitem__(self,key)

    def __setitem__(self, key, value):
        if not key is None:
            raise KeyError, "Undefined key %s for parameter %s" % (key, self.name)
        return _ParamBase.__setitem__(self,key,value)        

    def check_values(self):         #pragma:nocover
        #
        # Validate the values
        #
        if None not in self._index:
            raise ValueError, "Undefined value for parameter "+self.name
        if not self._valid_indexed_value(self.value,None,False):
            msg = "Parameter '%s' failed validation test.  Value: %s"
            raise ValueError, msg % ( self.name, str(self.value) )

    def __call__(self, exception=True):
        return self.value


class _ParamArray(_ParamBase):

    def __init__(self, *args, **kwds):
        _ParamBase.__init__(self, *args, **kwds)

    def __float__(self):
        msg = "Cannot access the value of array parameter '%s'"
        raise ValueError, msg % self.name

    def __int__(self):
        msg = "Cannot access the value of array parameter '%s'"
        raise ValueError, msg % self.name

    def set_value(self, value):
        for key in self._data:
            self.__setitem__(key,value)

    def check_values(self):         #pragma:nocover
        #
        # Validate the values
        #
        for key in self._data:
            if key not in self._index:
                msg = "Undefined value for parameter %s[%s]"
                raise ValueError, msg % ( self.name, str(key) )

            val = self._data(key)
            if not self._valid_indexed_value( val, key, False ):
                msg = "Parameter %s[%s] failed validation test.  Value: %s"
                raise ValueError, msg % ( self.name, str(key), str(tval) )

    def reset(self):
        pass                        #pragma:nocover

    def display(self, ostream=None):
        self.pprint(ostream=ostream)

    def __str__(self):
        return str(self.name)


class Param(Component):
    """
    Data objects that are used to construct Pyomo models.
    """

    alias("Param", "Parameter data that is used to define a model instance.")

    def __new__(cls, *args, **kwds):
        if args == ():
            self = _ParamElement(*args, **kwds)
        else:
            self = _ParamArray(*args, **kwds)
        return self



# Identical to example1.py except using linearizations
# at user given tangent points
#
#          / -X+2 , -5 <= X <= 1
#  Z(X) >= |
#          \  X ,  1 <= X <= 5

from coopr.pyomo import *

# Define the function
# Just like in Pyomo constraint rules, a Pyomo model object
# must be the first argument for the function rule
def f(model,x):
    return abs(x-1) + 1.0

# Define the derivative
def df(model,x):
    if x < 1:
        return -1
    else:
        return 1

model = ConcreteModel()

# Unlike example1.py, bounds on model.X are needed since
# linearizations don't provide outer bounds. In this 
# case the solution is at X=1 so the bounds aren't 
# really needed.
model.X = Var(bounds=(-5,5))
model.Z = Var()

# See documentation on Piecewise component by typing
# help(Piecewise) in a python terminal after importing coopr.pyomo
model.con = Piecewise(model.Z,model.X, # range and domain variables
                      tan_pts=[-5,5],
                      tan_constr_type='LB',
                      f_rule=f,
                      d_f_rule=df)

model.obj = Objective(expr=model.Z, sense=minimize)




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________


import pyutilib.component.core
from coopr.pyomo.base import expr, Constraint, Objective, SOSConstraint
from coopr.pyomo.base.connector import _ConnectorBase, _ConnectorValue
from coopr.pyomo.base.numvalue import NumericConstant
from coopr.pyomo.base.var import  _VarData, VarStatus, _VarArray, _VarBase, Var, VarList
from coopr.pyomo.base import IPyomoPresolver, IPyomoPresolveAction
import logging


class IdentifyVariablesPresolver(pyutilib.component.core.SingletonPlugin):
    """
    This plugin processes a Pyomo Model instance to define
    the order of variables and to categorize constraints.
    """

    pyutilib.component.core.implements(IPyomoPresolveAction)

    def __init__(self, **kwds):
        kwds['name'] = "identify_variables"
        pyutilib.component.core.Plugin.__init__(self, **kwds)

    def rank(self):
        return 1

    #
    # Scan through a linearly encoded expression body to identify variables.
    #
    def _identify_variables_in_linear_encoding(self, encoding):

        # recall that "encoding" is a tuple: element 0 is the constant, element 1 is a list of (coefficient, _VarData) pairs.
        for coefficient, var in encoding[1]:

            if var.fixed: # we know this is a _VarData, by construction - do the direct query to save time.
                continue
            else:
                var.status = VarStatus.used

    #
    # Recursively work through an expression (tree) to identify variables.
    #
    def _identify_variables_in_expression(self, exp):

        try:
            # It is possible (likely) that the expression is not
            # completely valid.  Check if that is the case, and return a
            # useful message to the user.
            is_expr = exp.is_expression()
        except AttributeError, e:
            # Likely caused by a constraint or an objective failing
            # to return a value.
            if isinstance(exp, _VarArray):
                msg = "Have you properly indexed '%s'?" % str(exp)
            else:
                msg = "Do all constraints and objectives return a value?"
            logging.getLogger('coopr.pyomo').error(
                "identifying variables in expression failed: %s\n%s\ntype=%s",
                str(e), msg, str(type(exp)) )
            raise

        #
        # Expression
        #
        if is_expr:
            #
            # Product Expressions store stuff differently
            #

            # NOTE: Avoid the considerable cost of recursion if we can avoid it.
            if exp.__class__ is expr._ProductExpression:
                for arg in exp._numerator:
                    if arg.is_constant():
                        pass
                    elif isinstance(arg,_VarData):
                        arg.status = VarStatus.used
                    else:
                        self._identify_variables_in_expression(arg)
                for arg in exp._denominator:
                    if arg.is_constant():
                        pass
                    else:
                       self._identify_variables_in_expression(arg)
            else:
                for arg in exp._args:
                    if arg.is_constant():
                        pass
                    elif isinstance(arg,_VarData):
                        arg.status = VarStatus.used
                    else:
                        self._identify_variables_in_expression(arg)
        #
        # Fixed Values (params, constants, fixed variables)
        #
        elif exp.is_constant():
            pass
        #
        # Variable Value
        #
        elif isinstance(exp,_VarData):
            exp.status = VarStatus.used

        #
        # Connector Value
        #
        elif isinstance(exp, _ConnectorValue):
            for var in exp.vars.itervalues():
                # FIXME: HACK: this is needed because preprocess is
                # called before the instance transformation plugins --
                # so for connectors attached to VarLists, the connector
                # is still pointing to an empty var list.  When
                # preprocess is called again after the transofrmation,
                # the expanded constraints will reference individual
                # _VarData.
                if type(var) is not VarList:
                    self._identify_variables_in_expression(var)
        #
        # Variable Base
        #
        elif isinstance(exp,_VarBase):
            # FIXME: [JDS] I am pretty sure that the only way to get here
            # will always raise the ValueError exception; the rest of
            # this logic is superfluous
            tmp = exp.as_numeric()
            if tmp is exp:
                raise ValueError("Variable %s appears in an expression without "
                                 "an index, but this variable is an indexed "
                                 "value." % ( exp.name, ))
            exp.status = VarStatus.used

        #
        # Connector Base
        #
        elif isinstance(exp,_ConnectorBase):
            raise ValueError("Connector %s appears in an expression without "
                             "an index, but this variable is an indexed "
                             "value." % exp.name)
        #
        # If not a constant, then this is an error
        #
        else:
            raise ValueError, "Unexpected expression type in identify_variables: " + str(type(exp))+" "+str(exp)

    def _preprocess_block(self, block):
        active_variables = block.active_components(Var)
        active_constraints = block.active_components(Constraint)
        active_objectives = block.active_components(Objective)

        active_constraints.update(block.active_components(SOSConstraint))

        #
        # Call the appropriate identify_variables to find the variables that
        # are actively used in the objective and constraints.
        #
        for name, objective in active_objectives.iteritems():
            for index, objective_data in objective._data.iteritems():
                if not objective_data.active:
                    continue
                try:
                    self._identify_variables_in_expression(objective_data.expr)
                except ValueError, err:
                    logging.getLogger('coopr.pyomo').error(
                        "Problem processing objective %s (index %s): %s" %
                        (str(name), str(index), str(err)) )
                    raise

                objective_data.id = self.objective_count
                self.objective_count += 1

        for name, constraint in active_constraints.iteritems():
            for index, constraint_data in constraint._data.iteritems():
                if not constraint_data.active:
                    continue
                try:
                    if constraint_data.lin_body is not None:
                        self._identify_variables_in_linear_encoding(constraint_data.lin_body)
                    else:
                        self._identify_variables_in_expression(constraint_data.body)
                except ValueError, err:
                    logging.getLogger('coopr.pyomo').error(
                        "Problem processing constraint %s (index %s):\n%s" %
                        ( str(name), str(index), str(err) ) )
                    raise
                self.constraint_count += 1

        #
        # count the number of variables in each type category.  we don't
        # do this within the _identify_variables_in_expression function
        # because the variable type is owned by the "parent" variable,
        # and the interrogation cost for variable type is comparatively
        # expensive - we don't want to do it for each variable value.
        #
        for var in active_variables.itervalues():
            self.num_binary_var += len(var.binary_keys())
            self.num_integer_var += len(var.integer_keys())
            self.num_real_var += len(var.continuous_keys())

    #
    # this pre-processor computes a number of attributes of model
    # components, in addition to models themselves. for _VarData,
    # _ConstraintData, and _ObjectiveData, the routine assigns the "id"
    # attribute. in addition, for _VarData, the routine sets the
    # "status" attribute - to used or unused, depending on whether it
    # appears in an active constraint/objective. on the Model object,
    # the routine populates the "statistics" (variable / constraint /
    # objective counts) attribute, and builds the _var, _con, and _obj
    # dictionaries of a Model, which map the id attributes to the
    # respective _VarData, _ConstraintData, or _ObjectiveData object.
    #
    # NOTE: if variables are fixed, then this preprocessor will flag
    #       them as unused and consequently not create entries for them
    #       in the model _var map.
    #
    def preprocess(self,model):
        """
        The main routine to perform the preprocess
        """
        self.var_count=0
        self.constraint_count=0
        self.objective_count=0
        self.num_binary_var = 0
        self.num_integer_var = 0
        self.num_real_var = 0

        model.statistics.number_of_binary_variables = 0
        model.statistics.number_of_integer_variables = 0
        model.statistics.number_of_continuous_variables = 0

        all_blocks = model.all_blocks()

        #
        # Until proven otherwise, all variables are unused.
        #
        for block in all_blocks:
            for var in block.active_components(Var).itervalues():
                for var_value in var._varval.itervalues():
                    var_value.status = VarStatus.unused

        for block in all_blocks:
            self._preprocess_block(block)

        #
        # assign the model statistics
        #

        model.statistics.number_of_variables = self.var_count
        model.statistics.number_of_constraints = self.constraint_count
        model.statistics.number_of_objectives = self.objective_count
        model.statistics.number_of_binary_variables = self.num_binary_var
        model.statistics.number_of_integer_variables = self.num_integer_var
        model.statistics.number_of_continuous_variables = self.num_real_var
        return model



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Var', 'VarList', 'VarStatus', '_VarBase', '_VarData']

import logging
import types
import sys
import weakref

from pyutilib.component.core import alias
import pyutilib.math
import pyutilib.misc
from pyutilib.enum import Enum

from numvalue import NumericValue, value, as_numeric
from set_types import BooleanSet, IntegerSet, Reals
from component import Component
from indexed_component import IndexedComponent
from misc import apply_indexed_rule
from misc import create_name
from sets import Set, _SetContainer
from coopr.pyomo.base.util import isfunctor

logger = logging.getLogger('coopr.pyomo')

VarStatus = Enum( 'undefined', 'fixed_by_presolve', 'fixed_by_optimizer',
                  'unused', 'used', )

class _VarData(NumericValue):
    """Holds the numeric value of a variable"""

    __slots__ = ['component','index','initial','active','lb','ub','fixed','status']

    # IMPT: If a specific argument is intended to be left unassigned, use the following default values in the constructor below:
    # name: None
    # domain: None

    def __init__(self, name, domain, component):
        """Constructor"""

        # IMPT: The following three lines are equivalent to calling the
        #       basic NumericValue constructor, i.e., as follows:
        #       NumericValue.__init__(self, name, domain, None, False)
        #       That particular constructor call takes a lot of time
        #       for big models, and is unnecessary because we're not
        #       validating any values.
        self.name = name
        self.domain = domain
        self.value = None

        # NOTE: the "name" attribute (part of the base NumericValue class) is
        #       typically something like some_var[x,y,z] - an easy-to-read
        #       representation of the variable/index pair.

        # NOTE: the following is presently set by the parent. arguably, we
        #       should provide keywords to streamline initialization.
        if component is None:
            self.component = None
        else:
            self.component = weakref.ref(component)
        self.index = None # the index of this variable within the "parent"

        # the default initial value for this variable.
        self.initial = None

        # is this variable an active component of the current model?
        self.active = True

        # IMPORTANT: in contrast to the initial value, the lower and upper
        #             bounds of a variable can in principle be either numeric constants,
        #             parameter values, or expressions - anything is OK, as long as
        #             invoking () is legal.
        self.lb = None
        self.ub = None

        self.fixed = False
        self.status = VarStatus.undefined

    def __str__(self):
        # the name can be None, in which case simply return "".
        if self.name is None:
            return ""
        else:
            return self.name

    #
    # IMPT: This class over-rides __getstate__ and __setstate__, due to the 'var' and 'component' weakrefs, and because of slots.
    #

    def __getstate__(self):

        result = NumericValue.__getstate__(self)
        for i in _VarData.__slots__:
            result[i] = getattr(self, i)
        if type(result['component']) is weakref.ref:
            result['component'] = result['component']()
        return result

    def __setstate__(self, state):
        for (slot_name, value) in state.iteritems():
            setattr(self, slot_name, value)
        if self.component is not None:
            self.component = weakref.ref(self.component)

    def getattrvalue(self, attr_name):
        if attr_name not in self.component()._attr_values:
            raise RuntimeError,"Variable="+str(self.component().name)+" does not have an attribute defined with name="+str(attr_name)
        return self.component()._attr_values[attr_name][self.index]

    def setattrvalue(self, attr_name, attr_value, define=False):
        if attr_name not in self.component()._attr_values:
            if define:
                self.component().declare_attribute(attr_name)
            else:
                raise RuntimeError,"Variable="+str(self.component().name)+" does not have an attribute defined with name="+str(attr_name)
        self.component()._attr_values[attr_name][self.index] = attr_value

    def activate(self):
        self.active=True

    def deactivate(self):
        self.active=False

    def setlb(self, value):
        # python only has 3 numeric built-in type that we deal with (we ignore
        # complex numbers).
        if value is None:
            self.lb = None
        else:
            try:
                tmp = as_numeric(value)
                if tmp.is_constant():
                    self.lb = tmp
                else:
                    raise ValueError(
                        "Non-constant type '%s' supplied as variable lower "
                        "bound - legal types are numeric constants or "
                        "parameters"
                        % (type(value),) )
            except:
                raise ValueError(
                    "Unknown type '%s' supplied as variable lower bound - "
                    "legal types are numeric constants or parameters"
                    % (type(value),) )

    def setub(self, value):
        # python only has 3 numeric built-in type that we deal with (we ignore
        # complex numbers).
        if value is None:
            self.ub = None
        else:
            try:
                tmp = as_numeric(value)
                if tmp.is_constant():
                    self.ub = tmp
                else:
                    raise ValueError(
                        "Non-constant type '%s' supplied as variable upper "
                        "bound - legal types are numeric constants or "
                        "parameters"
                        % (type(value),) )
            except:
                raise ValueError(
                    "Unknown type '%s' supplied as variable upper bound - "
                    "legal types are numeric constants or parameters"
                    % (type(value),) )


    def fixed_value(self):
        if self.fixed:
            return True
        return False

    def is_constant(self):
        if self.fixed:
            return True
        return False

    def polynomial_degree(self):
        if self.fixed:
            return 0
        return 1

    def is_binary(self):
        return self.component().is_binary(self.index)

    def is_integer(self):
        return self.component().is_integer(self.index)

    def is_continuous(self):
        return self.component().is_continuous(self.index)

    def pprint(self, ostream=None, verbose=False):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, str(self),

    def _tighten_bounds(self):
        """
        Attempts to tighten the lower and upper bounds on
        this variable by using the domain.
        """
        try:
            dlb, dub = self.domain.bounds()
        except:
            # No explicit domain bounds
            return

        if (dlb is not None):
            if (self.lb is None) or (dlb > self.lb.value):
                self.setlb(dlb)

        if (dub is not None):
            if (self.ub is None) or (dub < self.ub.value):
                self.setub(dub)


#
# Variable attributes:
#
# Single variable:
#
# x = Var()
# x.fixed = True
# x.domain = Reals
# x.name = "x"
# x.setlb(-1.0)
# x.setub(1.0)
# x.initial = 0.0
# x = -0.4
#
# Array of variables:
#
# y = Var(set1,set2,...,setn)
# y.fixed = True (fixes all variables)
# y.domain = Reals
# y.name = "y"
# y[i,j,...,k] (returns value of a given index)
#
# y[i,j,...,k].fixed
# y[i,j,...,k].lb()
# y[i,j,...,k].ub()
# y[i,j,...,k].initial
#

class _VarBase(IndexedComponent):
    """A numeric variable, which may be defined over a index"""

    """ Constructor
        Arguments:
           name         The name of this variable
           index        The index set that defines the distinct variables.
                          By default, this is None, indicating that there
                          is a single variable.
           domain       A set that defines the type of values that
                          each variable must be.
           default      A set that defines default values for this
                          variable.
           bounds       A rule for defining bounds values for this
                          variable.
           rule         A rule for setting up this variable with
                          existing model data
    """

    #
    # IMPT: The default __getstate__ and __setstate__ are fine - this class owns no weakrefs or slots,
    #       and should be able to rely on the base IndexedComponent methods.
    #

    # TODO: default and rule keywords are not used?  Need to talk to Bill ...?
    def __init__(self, *args, **kwd):

        # Default keyword values
        self._initialize = kwd.pop('initialize', None )
        self.domain = kwd.pop('within', Reals )
        self.domain = kwd.pop('domain', self.domain )
        self.bounds = kwd.pop('bounds', None )

        self._defer_domain = False

        # the guts of the class - maps indicies to _VarData objects.
        self._varval = {}

        self._attr_declarations = {} # maps the name of an attribute (i.e., a suffix) to its default value.
        self._attr_values = {} # maps the name of an attribute to a dictionary containing (index, value) pairs.

        #Handle within=not_pyomo_object 
        if not hasattr(self.domain,'virtual') or not self.domain.virtual:
            if self.domain == [] or self.domain == set([]):
                raise ValueError,"Attempting to set a variable's domain to an empty set"
            else: 
                from rangeset import RangeSet
                if type(self.domain) is RangeSet:
                    self.bounds = (self.domain._start,self.domain._end)
                    domain_name = self.domain.name
                    self.domain = IntegerSet()
                    self.domain.name = domain_name
                elif type(self.domain) is _SetContainer:
                    if self.domain.initialize is None:
                        self._defer_domain = True
                    elif self.domain.initialize == []:
                        raise ValueError,"Attempting to set a variable's domain to an empty Set"
                    else:
                        self._defer_domain = False
                        domain_name = self.domain.name
                        self.domain = self.domain.initialize
                        self.domain.sort()
                        if len(self.domain) != self.domain[-1] - self.domain[0] + 1:
                            raise ValueError,"Attempting to set a variable's domain to an improperly formatted Set " \
                                           + "-- elements are missing or duplicates exist: {0}".format(self.domain)
                        elif not all(type(elt) is int for elt in self.domain):
                            raise ValueError,"Attempting to set a veriable's domain to a Set with noninteger elements: {0}".format(self.domain)
                        else:
                            self.bounds = (self.domain[0],self.domain[-1])
                            self.domain = IntegerSet()
                            self.domain.name = domain_name
                elif type(self.domain) is list:
                    self.domain.sort()
                    if len(self.domain) != self.domain[-1] - self.domain[0] + 1:
                        raise ValueError,"Attempting to set a variable's domain to an improperly formatted set " \
                                       + "-- elements are missing or duplicates exist: {0}".format(self.domain)
                    elif not all(type(elt) is int for elt in self.domain):
                        raise ValueError,"Attempting to set a veriable's domain to a set with noninteger elements: {0}".format(self.domain)
                    else:
                        self.bounds = (self.domain[0],self.domain[-1])
                        domain_name = "[{0}..{1}]".format(self.domain[0],self.domain[-1])
                        self.domain = IntegerSet()
                        self.domain.name = domain_name
                elif type(self.domain) is set:
                    self.domain = list(self.domain) 
                    self.domain.sort()
                    if len(self.domain) != self.domain[-1] - self.domain[0] + 1:
                        raise ValueError,"Attempting to set a variable's domain to an improperly formatted set " \
                                       + "-- elements are missing or duplicates exist: {0}".format(self.domain)
                    elif not all(type(elt) is int for elt in self.domain):
                        raise ValueError,"Attempting to set a veriable's domain to a set with noninteger elements: {0}".format(self.domain)
                    else:
                        self.bounds = (self.domain[0],self.domain[-1])
                        domain_name = "[{0}..{1}]".format(self.domain[0],self.domain[-1])
                        self.domain = IntegerSet()
                        self.domain.name = domain_name
                elif type(self.domain) is xrange:
                    self.bounds = (self.domain[0],self.domain[-1])
                    domain_name = "[{0}..{1}]".format(self.bounds[0],self.bounds[-1])
                    self.domain = IntegerSet()
                    self.domain.name = domain_name

        kwd.setdefault('ctype', Var)

        IndexedComponent.__init__(self, *args, **kwd)

        # check to see if domains for individual _VarData objects are specified via a rule.
        if isfunctor(self.domain):
            self._domain_rule = self.domain
            self.domain = None # there is no general domain for the collection of variables.
        else:
            self._domain_rule = None # there is no domain rule!

        # cached for efficiency purposes - isinstance is not cheap. 
        # only truly valid if there is no domain rule specified.
        if self.domain is not None:
            self._is_binary = isinstance(self.domain, BooleanSet)
            # second case is required in the case of deferred domains (the set may not be constructed yet, so hasn't been translated to an IntegerSet)
            self._is_integer = isinstance(self.domain, IntegerSet) or (type(self.domain) is _SetContainer)
            self._is_continuous = not (self._is_binary or self._is_integer)
        else:
            self._is_binary = None
            self._is_integer = None
            self._is_continuous = None

        # the following are only applicable if a domain rule *has* been specified.
        self._binary_keys = []
        self._integer_keys = []
        self._continuous_keys = []

    def as_numeric(self):
        if None in self._varval:
            return self._varval[None]
        return self

    def is_indexed(self):
        return self._ndim > 0

    def is_expression(self):
        return False

    def is_relational(self):
        return False

    def keys(self):
        return self._varval.keys()

    # returns a dictionary of index-value pairs.
    def extract_values(self):
        return dict(((index, varval.value) for index, varval in self._varval.iteritems()))

    # takes as input a dictionary of index-value pairs.
    def store_values(self, new_values):
        for index, varval in self._varval.iteritems():
            varval.value = new_values[index]

    def is_binary(self, index):
        if self._domain_rule is None:
            return self._is_binary
        else:
            return index in self._binary_keys

    def is_integer(self, index):
        if self._domain_rule is None:
            return self._is_integer
        else:
            return index in self._integer_keys

    def is_continuous(self, index):
        if self._domain_rule is None:
            return self._is_continuous
        else:
            return index in self._continuous_keys

    def binary_keys(self):
        """ Returns the keys of all binary variables """
        if self._domain_rule is None:
            if self._is_binary:
                return self._varval.keys()
            else:
                return []
        else:
            return self._binary_keys

    def integer_keys(self):
        """ Returns the keys of all integer variables """
        if self._domain_rule is None:
            if self._is_integer:
                return self._varval.keys()
            else:
                return []
        else:
            return self._integer_keys

    def continuous_keys(self):
        """ Returns the keys of all continuous variables """
        if self._domain_rule is None:
            if self._is_continuous:
                return self._varval.keys()
            else:
                return []
        else:
            return self._continuous_keys

    def __iter__(self):
        return self._varval.iterkeys()

    def iterkeys(self):
        return self._varval.iterkeys()

    def itervalues(self):
        return self._varval.itervalues()

    def iteritems(self):
        return self._varval.iteritems()

    def __contains__(self, ndx):
        return ndx in self._varval

    def dim(self):
        return self._ndim

    def reset(self):
        for value in self._varval.itervalues():
            value.set_value(value.initial)

    def __len__(self):
        return len(self._varval)

    def __setitem__(self,ndx,val):
        #print "HERE",ndx,val, self._valid_value(val,False), self.domain # XXX debugging
        if (len(self._index) == 1) and (None in self._index) and (ndx is not None): # a somewhat gory way to see if you have a singleton - the len()=1 check is needed to avoid "None" being passed illegally into set membership validation rules.
            # allow for None indexing if this is truly a singleton
            msg = "Cannot set an array value in singleton variable '%s'"
            raise KeyError, msg % self.name

        if ndx not in self._index:
            msg = "Cannot set the value of array variable '%s' with invalid " \
                  "index '%s'"
            raise KeyError, msg % ( self.name, str(ndx) )

        if not self._valid_value(val,False):
            msg = "Cannot set variable '%s[%s]' with invalid value: %s"
            raise ValueError, msg % ( self.name, str(ndx), str(val) )
        self._varval[ndx].value = val

    def __getitem__(self,ndx):
        """This method returns a _VarData object.  This object can be
           coerced to a numeric value using the value() function, or using
           explicity coercion with float().
        """
        try:
            return self._varval[ndx]
        except KeyError: # thrown if the supplied index is hashable, but not defined.
            msg = "Unknown index '%s' in variable %s;" % (str(ndx), self.name)
            if (isinstance(ndx, (tuple, list)) and len(ndx) != self.dim()):
                msg += "    Expecting %i-dimensional indices" % self.dim()
            else:
                msg += "    Make sure the correct index sets were used.\n"
                msg += "    Is the ordering of the indices correct?"
            raise KeyError, msg
        except TypeError, msg: # thrown if the supplied index is not hashable
            msg2 = "Unable to index variable %s using supplied index with " % self.name
            msg2 += str(msg)
            raise TypeError, msg2

    def _add_indexed_member(self, ndx):
        # Check for domain rules
        if self._domain_rule is not None:
            domain = apply_indexed_rule( self, self._domain_rule,
                                         self._model(), ndx )
            if isinstance(domain, BooleanSet):
                self._binary_keys.append(ndx)
            elif isinstance(domain, IntegerSet):
                self._integer_keys.append(ndx)
            else:
                self._continuous_keys.append(ndx)
        else:
            domain = self.domain

        new_varval = _VarData(create_name(self.name,ndx), domain, self)
        new_varval.index = ndx
        
        self._varval[ndx] = new_varval


    def construct(self, data=None):
        if __debug__:   #pragma:nocover
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing Variable, name=%s, from data=%s", self.name, `data`)
        if self._constructed:
            return
        self._constructed=True
        #
        # Construct _VarData objects for all index values
        #
        if self._ndim > 0:
            
            if type(self._initialize) is dict:
                self._index = self._initialize.keys()

            for ndx in self._index:
                # TODO: it is common to construct empty VarLists... and
                # when that happens, the empty implicit set comes in
                # with a "None" index!
                if ndx is not None: 
                    self._add_indexed_member(ndx)


        #
        # Define the _XXX_keys objects if domain isn't a rule;
        # they were defined individually above in the case of
        # a rule
        #
        if self._domain_rule is None:
            if isinstance(self.domain, BooleanSet):
                self._binary_keys = self._varval.keys()
            elif isinstance(self.domain, IntegerSet):
                self._integer_keys = self._varval.keys()
            else:
                self._continuous_keys = self._varval.keys()

        #
        # Initialize values with a dictionary if provided
        #
        if self._initialize is not None:
            #
            # Initialize values with the _rule function if provided
            #
            if self._initialize.__class__ is types.FunctionType:
                for key in self._varval:
                    if key is None:
                        val = self._initialize(self._model())
                    else:
                        val = apply_indexed_rule( self, self._initialize,
                                                  self._model(), key )
                    val = value(val)
                    self._valid_value(val, True)
                    self._varval[key].value = self._varval[key].initial = val
            elif self._initialize.__class__ is dict:
                for key in self._initialize:
                    val = self._initialize[key]
                    self._valid_value(val, True)
                    self._varval[key].value = self._varval[key].initial = val
            else:
                for key in self._index:
                    val = self._initialize
                    self._valid_value(val, True)
                    self._varval[key].value = self._varval[key].initial = val
        #
        # Initialize bounds with the bounds function if provided
        #
        if self.bounds is not None:

            if type(self.bounds) is tuple:

                # bounds are specified via a tuple - same lower and upper bounds for all var values!

                (lb, ub) = self.bounds

                # do some simple validation that the bounds are actually finite - otherwise, set them to None.
                if (lb is not None) and (not pyutilib.math.is_finite(value(lb))):
                    lb = None
                if (ub is not None) and (not pyutilib.math.is_finite(value(ub))):
                    ub = None

                for key,varval in self._varval.iteritems():
                    if lb is not None:
                        varval.setlb(lb)
                    if ub is not None:
                        varval.setub(ub)

            else:

                # bounds are specified via a function

                for index, varval in self._varval.iteritems():
                    if index is None:
                        (lb, ub) = self.bounds(self._model())
                    else:
                        (lb, ub) = apply_indexed_rule( self, self.bounds,
                                                       self._model(), index )
                    varval.setlb(lb)
                    varval.setub(ub)

                for varval in self._varval.itervalues():
                    if varval.lb is not None and not pyutilib.math.is_finite(varval.lb()):
                        varval.setlb(None)
                    if varval.ub is not None and not pyutilib.math.is_finite(varval.ub()):
                        varval.setub(None)
        #
        # Iterate through all variables, and tighten the bounds based on
        # the domain bounds information.
        #
        # Only done if self.domain is not a rule. If it is, _VarArray level
        # bounds become meaningless, since the individual _VarElement objects
        # likely have more restricted domains.
        #
        if self._domain_rule is None:
            dbounds = self.domain.bounds()
            if not dbounds is None and dbounds != (None,None):
                for varval in self._varval.itervalues():
                    if not dbounds[0] is None:
                        if varval.lb is None or                     \
                           dbounds[0] > varval.lb():
                            varval.setlb(dbounds[0])

                    if not dbounds[1] is None:
                        if varval.ub is None or                     \
                            dbounds[1] < varval.ub():
                            varval.setub(dbounds[1])
        #
        # Setup declared attributes for all variables
        #
        for attr_name in self._attr_declarations:
            default_value = self._attr_declarations[attr_name][0]
            suffix_dict = {}
            for key in self._varval:
                suffix_dict[key]=default_value
            self._attr_values[attr_name] = suffix_dict

    def pprint(self, ostream=None, verbose=False):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, "  ",self.name,":",
        if not self.doc is None:
            print >>ostream, self.doc
            print >>ostream, "  ",
        print >>ostream, "\tSize="+str(len(self)),
        if self.domain is not None:
            print >>ostream, "\tDomain="+self.domain.name
        else:
            print >>ostream, "\tDomain=None"
        if self._index_set is not None:
            print >>ostream, "\tIndicies: ",
            for idx in self._index_set:
                print >>ostream, str(idx.name)+", ",
            print ""
        if None in self._varval:
            print >>ostream, "\tInitial Value : Lower Bound : Upper Bound : "  \
                             "Current Value: Fixed: Status"
            lb_value = None
            if self._varval[None].lb is not None:
                lb_value = self._varval[None].lb()
            ub_value = None
            if self._varval[None].ub is not None:
                ub_value = self._varval[None].ub()

            print >>ostream, "\t %s : %s : %s : %s : %s : %s" % (
              str(self._varval[None].initial),
              str(lb_value),
              str(ub_value),
              str(self._varval[None].value),
              str(self._varval[None].fixed),
              str(self._varval[None].status)
            )
        else:
            print >>ostream, "\tKey : Initial Value : Lower Bound : "          \
                             "Upper Bound : Current Value: Fixed: Status"
            tmp=self._varval.keys()
            tmp.sort()
            for key in tmp:
                initial_val = self._varval[key].initial
                lb_value = None
                if self._varval[key].lb is not None:
                    lb_value = self._varval[key].lb()
                ub_value = None
                if self._varval[key].ub is not None:
                    ub_value = self._varval[key].ub()

                print >>ostream, "\t%s : %s : %s : %s : %s : %s : %s" % (
                  str(key),
                  str(initial_val),
                  str(lb_value),
                  str(ub_value),
                  str(value(self._varval[key].value)),
                  str(value(self._varval[key].fixed)),
                  str(self._varval[key].status)
                )


    def display(self, prefix="", ostream=None):
        if ostream is None:
            ostream = sys.stdout
        print >>ostream, prefix+"Variable "+self.name,":",
        print >>ostream, "  Size="+str(len(self)),
        print >>ostream, "Domain="+self.domain.name
        if None in self._varval:
            print >>ostream, "%s  Value=%s" % (
              prefix,
              pyutilib.misc.format_io(self._varval[None].value)
            )
        else:
            for key in self._varval:
                val = self._varval[key].value
                print >>ostream, prefix+"  "+str(key)+" : "+str(val)

    def declare_attribute(self, name, default=None):
        """
        Declare a user-defined attribute.
        """
        if name[0] == "_":
            msg = "Cannot define an attribute that begins with  '_'"
            raise AttributeError, msg
        if name in self._attr_declarations:
            raise AttributeError, "Attribute %s is already defined" % name
        self._attr_declarations[name] = (default,)
        #if not default is None:
            #self._valid_value(default)
        #
        # If this variable has been constructed, then
        # generate this attribute for all variables
        #
        # add to self._attr_values
        if len(self._varval) > 0:
            suffix_dict = {}
            for key in self._varval:
                suffix_dict[key]=default
            self._attr_values[name] = suffix_dict

    def attribute_defined(self, name):
        """
        Determine if a user-defined attribute actually exists!
        """
        return name in self._attr_declarations

# a _VarElement is the implementation representing a "singleton" or non-indexed variable.
# NOTE: this class derives from both a "slot"ized base class (_VarData) and a normal
#       class with a dictionary (_VarBase) - beware (although I believe the __getstate__
#       implementation should be attribute-independent).

class _VarElement(_VarBase, _VarData):

    def __init__(self, *args, **kwd):

        _VarData.__init__(self,
                            kwd.get('name', None), 
                            kwd.get('within', kwd.get('domain', Reals)), 
                            self)
        _VarBase.__init__(self, *args, **kwd)
        self._varval[None] = self
        self._varval[None].index = None

    def __call__(self, exception=True):
        if None in self._varval:
            return self._varval[None].value
        return None

    def __getstate__(self):

        # picks up all slots in the _VarData class, and in particular
        # makes sure the "var" and "component" values are not weakrefs,
        # but the actual objects themselves.

        varvalue_result = _VarData.__getstate__(self)
        varvalue_result['component'] = self
        varbase_result = _VarBase.__getstate__(self)
        return dict(varvalue_result.items() + varbase_result.items())

    def __setstate__(self, state):

        for (slot_name, value) in state.iteritems():
            setattr(self, slot_name, value)

        self._varval[None].component = weakref.ref(self)

        # this is hackish, in that it replicates code from the base Component setstate method - revisit at some point.
        if '_parent' in self.__dict__.keys() and self._parent is not None and type(self._parent) != weakref.ref:
            self._parent = weakref.ref(self._parent)

    def is_constant(self):
        return _VarData.is_constant(self)

# a _VarArray is the implementation representing an indexed variable.

class _VarArray(_VarBase):
    
    def __init__(self, *args, **kwds):
        _VarBase.__init__(self, *args, **kwds)
        self._dummy_val = _VarData(kwds.get('name', None), 
                                    kwds.get('within', kwds.get('domain', Reals)), 
                                    self)

    def __float__(self):
        raise TypeError, "Cannot access the value of array variable "+self.name

    def __int__(self):
        raise TypeError, "Cannot access the value of array variable "+self.name

    def _valid_value(self,value,use_exception=True):
        return self._dummy_val._valid_value(value, use_exception)

    def set_value(self, value):
        msg = "Cannot specify the value of array variable '%s'"
        raise ValueError, msg % self.name

    def __str__(self):
        return self.name

    def construct(self, data=None):
        _VarBase.construct(self, data)

        for (ndx, var) in self._varval.iteritems():
            var._tighten_bounds()


class Var(Component):
    """
    Variable objects that are used to construct Pyomo models.
    """

    alias("Var", "Decision variables in a model.")

    def __new__(cls, *args, **kwds):
        if args == ():
            self = _VarElement(*args, **kwds)
        else:
            self = _VarArray(*args, **kwds)
        return self


class VarList(_VarArray):
    """
    Variable-length indexed variable objects used to construct Pyomo models.
    """

    alias("VarList", "Variable-length list of decision variables in a model.")

    End = ( 1003, )

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            raise ValueError, "Cannot specify indices for a VarList object"
        # Construct parents...
        self._hidden_index = Set()
        _VarArray.__init__(self, self._hidden_index, **kwargs)
        self._ndim = 1
        self._nvars = 0

    def fixed_value(self):
        if self._nvars == 0:
            return False
        for idx in xrange(self._nvars):
            if not self[idx].fixed_value():
                return False
        return True

    def is_constant(self):
        return self.fixed_value()

    def polynomial_degree(self):
        if self.is_constant():
            return 0
        return 1

    def construct(self, data=None):
        if __debug__:
            logger.debug("Constructing variable list %s",self.name)
        self._hidden_index.construct()
        _VarArray.construct(self, data)

    def add(self):
        self._hidden_index.add(self._nvars)
        self._add_indexed_member(self._nvars)
        self._nvars += 1
        return self[self._nvars-1]



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['_ParamData']

import weakref

from pyutilib.component.core import *
from pyutilib.misc import SparseMapping
import pyutilib.math

from set_types import Reals
from numvalue import NumericConstant
from plugin import IParamRepresentation
from misc import create_name


class _ParamData(NumericConstant):
    """Holds the numeric value of a parameter"""

    __slots__ = []

    def __init__(self, name, domain, value):
        """Constructor"""
        #
        # NumericConstant.__init__(self, name, domain, value)
        #
        if domain is None:
            self.domain = Reals
        else:
            self.domain = domain
        self.name = name
        if pyutilib.math.is_nan(value):
            self.value = pyutilib.math.nan
        else:
            self.value = value


    def __str__(self):
        return str(self.name)


class SparseMappingRepn(Plugin):

    implements(IParamRepresentation, service=False)

    alias('sparse_dict', 'A class that manages the sparse dictionary parameter representation, which maps keys to numeric values while respecting default values')

    def __init__(self, param):
        self._paramdata = SparseMapping(default=param._default.value, index=param._index, within=param.domain)
        self._param = weakref.ref(param)

    def model_repn(self):
        if None in self.keys():
            return self._paramdata[None]
        return self._paramdata

    def set_default(self, value):
        self._paramdata.default = value

    def update_index(self):
        self._paramdata._index = self._param()._index

    def data(self):
        return self._paramdata

    def keys(self, nondefault=False):
        if nondefault:
            self._paramdata.nondefault_keys()
        return self._paramdata.keys()

    #
    # a utiliy to extract all index-value pairs defining this
    # parameter, returned as a dictionary. useful in many
    # contexts, in which key iteration and repeated __getitem__
    # calls are too expensive to extract the contents of a parameter.
    # NOTE: We are presently not careful if an index has not been
    #       explicitly defined - issues around defaults are bound
    #       to crop up. specifically, we're assuming here that
    #       all indices are represented in _paramdata.
    #
    def extract_values(self):
        raise RuntimeError, "extract_values method not yet implemented for SparseMappingRepn"

    def store_values(self, new_values):
        raise RuntimeError, "store_values method not yet implemented for SparseMappingRepn"    

    def __contains__(self, key):
        return key in self._paramdata

    def __iter__(self):
        return self._paramdata.__iter__()

    def __len__(self):
        return len(self._paramdata)

    def __getitem__(self, key):
        return self._paramdata[key]

    def __setitem__(self, key, val):
        self._paramdata[key] = val

    def set_item(self, key=None, val=None):
        if key is None:
            self._param()._default.value = val
            self._paramdata = SparseMapping(default=self._param()._default.value, index=self._param()._index, within=self._param().domain)
        else:
            self._paramdata.set_item(key,val)

    def __call__(self, key):
        return self._paramdata[key]

    def __getstate__(self):
        """
        This method must be defined to support pickling because this class
        owns a weakref for '_param'.
        """
        result = dict(self.__dict__.items())
        if ('_param' in result) and (type(result['_param']) is weakref.ref):
            result['_param'] = result['_param']()
        return result

    def __setstate__(self, state):
        """
        This method must be defined to support pickling because this class
        owns a weakref for '_param'.
        """
        for (slot_name, value) in state.iteritems():
            self.__dict__[slot_name] = value
        if '_param' in self.__dict__.keys() and self._param is not None:
            self._param = weakref.ref(self._param)


class DefaultRepn(Plugin):

    implements(IParamRepresentation, service=False)

    alias('pyomo_dict', 'A class that manages the default Pyomo parameter representation, which maps keys to _ParamData objects')

    def __init__(self, param):
        self._paramdata = {} # mapping between indicies and _ParamData objects.
        self._default = param._default
        self._index = param._index
        self._param = weakref.ref(param)

    def model_repn(self):
        return self._param()

    def set_default(self, value):
        pass

    def update_index(self):
        self._index = self._param()._index

    def data(self):
        tmp = SparseMapping(default=self._default.value, index=self._param()._index, within=self._param().domain)
        for key in self._paramdata.keys():
            tmp[key] = self._paramdata[key].value
        return tmp

    def keys(self, nondefault=False):
        if nondefault or self._default.value is None:
            return self._paramdata.keys()
        return self._index

    # a utiliy to extract all index-value pairs defining this
    # parameter, returned as a dictionary. useful in many
    # contexts, in which key iteration and repeated __getitem__
    # calls are too expensive to extract the contents of a parameter.
    # NOTE: We are presently not careful if an index has not been
    #       explicitly defined - issues around defaults are bound
    #       to crop up. specifically, we're assuming here that
    #       all indices are represented in _paramdata.
    def extract_values(self):
        return dict(((index, paramval.value) for index, paramval in self._paramdata.iteritems()))

    # a utility to set values of a parameter "in bulk", using a dictionary of index<->value pairs.
    # far more efficient in some contexts than the approach of doing things by-index in an outer loop.
    def store_values(self, new_values):
        for index, new_value in new_values.iteritems():
            self._paramdata[index].value = new_value            

    def __contains__(self, key):
        if self._default.value is None:
            return key in self._paramdata
        else:
            return key in self._index

    def __iter__(self):
        if self._default.value is None:
            return self._paramdata.keys().__iter__()
        else:
            return self._index.__iter__()

    def __len__(self):
        if self._default.value is None:
            return len(self._paramdata)
        return len(self._index)

    def __getitem__(self, key):
        if key in self._paramdata:
            return self._paramdata[key]
        if key in self._index:
            if self._default.value is not None:
                return self._default
            else: 
                raise ValueError( "Legal key '%s' for parameter '%s', but "
                                  "value is uninitialized and there is no "
                                  "default value" % (key, self._param().name) )
        raise KeyError( "Undefined key '%s' for parameter '%s'"
                        % (key, self._param().name) )

    def __setitem__(self,key,val):
        if key in self._paramdata:
            self._paramdata[key].value = val
        else:
            self._paramdata[key]= _ParamData(
                        create_name(self._param().name,key),
                        self._param().domain,
                        val
                        )

    def set_item(self, key=None, val=None):
        if key is None:
            self._paramdata = {}
            self._default.value = getattr(val,'value',val)
        else:
            self._paramdata[key] = val

    def __call__(self, key):
        return self.__getitem__(key).value

    def __getstate__(self):
        """
        This method must be defined to support pickling because this class
        owns a weakref for '_param'.
        """
        result = dict(self.__dict__.items())
        if ('_param' in result) and (type(result['_param']) is weakref.ref):
            result['_param'] = result['_param']()
        return result

    def __setstate__(self, state):
        """
        This method must be defined to support pickling because this class
        owns a weakref for '_param'.
        """
        for (slot_name, value) in state.iteritems():
            self.__dict__[slot_name] = value
        if '_param' in self.__dict__.keys() and self._param is not None:
            self._param = weakref.ref(self._param)




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Model', 'ConcreteModel', 'AbstractModel']

import array
import copy
import logging
import re
import sys
import traceback
import weakref
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from plugin import *
from numvalue import *

import pyutilib.component.core
from pyutilib.math import *
from pyutilib.misc import quote_split, tuplize, Container, PauseGC

from coopr.opt import ProblemFormat, ResultsFormat, guess_format
from coopr.opt.base import SymbolMap
from coopr.opt.results import SolutionMap, SolverResults, Solution
from coopr.pyomo.base.var import _VarData, VarStatus, Var
from coopr.pyomo.base.constraint import _ConstraintData
from coopr.pyomo.base.set_types import *
import coopr.opt
from PyomoModelData import ModelData
from coopr.opt.results.container import MapContainer,UndefinedData

from block import Block
from sets import Set

try:
    from pympler.muppy import tracker
    pympler_available = True
except ImportError:
    pympler_available = False

logger = logging.getLogger('coopr.pyomo')

class Model(Block):
    """
    An optimization model.  By default, this defers construction of components
    until data is loaded.
    """

    pyutilib.component.core.alias("Model", 'Model objects can be used as a '  \
                                  'component of other models.')

    preprocessor_ep = pyutilib.component.core.ExtensionPoint(IPyomoPresolver)

    #
    # IMPT: The base Block __getstate__ and __setstate__ methods are fine - 
    #       this class owns no weakrefs or slots.
    #

    def __init__ ( self, name='unknown', _deprecate=True, **kwargs ):
        """Constructor"""
        if _deprecate:
            msg = "Using the 'Model' class is deprecated.  Please use the "    \
                  "'AbstractModel' class instead."
            logger.warning( msg )

        # NB: some folks (notably, the PySP EF generator will stick
        # sub-models (the scenario models) into a master model (the
        # binding instance), so life is infinitely easier if the Model
        # registers itself as a Block -- otherwise, we need to define
        # Block.all_blocks() to look for both Blocks and Models; which
        # creates a chicken & egg import problem (block.py needs the
        # definition of Model for all_blocks(), but PyomoModel.py needs
        # to import block.py so inheritance works!
        #
        #kwargs.setdefault('ctype',Model)
        Block.__init__(self, **kwargs)

        self.name=name
        self._preprocessors = ["simple_preprocessor"]
        #
        # Model statistics
        #
        self.statistics = Container()

    def nvariables(self):
        return self.statistics.number_of_variables

    #def variable(self, name):
    #    if isinstance(name,basestring):
    #        return self._label_var_map[name]
    #    else:
    #        return self._var[name-1]

    #def variables(self):
    #    return self._label_var_map

    def nconstraints(self):
        return self.statistics.number_of_constraints

    #def constraint(self, name):
    #    if isinstance(name,basestring):
    #        return self._label_constraint_map[name]
    #    else:
    #        return self._con[name-1]

    #def constraints(self):
    #    return self._label_constraint_map

    def nobjectives(self):
        return self.statistics.number_of_objectives

    #def objective(self, name):
    #    if isinstance(name,basestring):
    #        return self._label_objective_map[name]
    #    else:
    #        return self._obj[name-1]

    #def objectives(self):
    #    return self._label_objective_map

    #def num_used_variables(self):
    #    return len(self._var)

    def valid_problem_types(self):
        return [ProblemFormat.pyomo]

    def create(self, filename=None, name=None, namespace=None, namespaces=None, preprocess=True, simplify=None, profile_memory=0):
        """
        Create a concrete instance of this Model, possibly using data
        read in from a file.
        """

        if simplify is not None:
            print """
 WARNING: The 'simplify' argument to PyomoModel.create() has been deprecated.
      Please remove references from your code.
 """

        # if you have a concrete model, it's already constructed - passing in a
        # data file is a waste of time. the only reason one would call create()
        # on a concrete instance is to preprocess.
        if (self._defer_construction is False) and isinstance(filename,basestring) and filename is not None:
            name_to_print = self.name
            if name is not None:
                name_to_print = name
            msg = "The filename=%s will not be loaded - supplied as an argument to the create() method of a ConcreteModel instance with name=%s." % (filename, name)
            logger.warning(msg)
            
        if self._defer_construction:
            instance = self.clone()

            if namespaces is None:
                instance.load(filename, namespaces=[namespace], profile_memory=profile_memory)
            elif len(namespaces) == 0:
                instance.load(filename, namespaces=[None], profile_memory=profile_memory)
            else:
                instance.load(filename, namespaces=namespaces+[None], profile_memory=profile_memory)
        else:
            instance = self
            
        # FIXME: Logically, we should do the following, but if I do, it
        # breaks several set tests
        #
        #instance._defer_construction = False
        #instance._constructed = True

        if preprocess is True:

            if (pympler_available is True) and (profile_memory >= 2):
                memory_tracker = tracker.SummaryTracker()
                memory_tracker.create_summary()

            instance.preprocess()

            if (pympler_available is True) and (profile_memory >= 2):
                print "Objects created during instance preprocessing:"
                memory_tracker.print_diff()

        if not name is None:
            instance.name=name
        return instance

    def clone(self):
        """Create a copy of this model"""

        #for name in self.components:
        #    #print 'y',name
        #    #print 'z', type(self.components[name])
        #    self.components[name]._model = None
        self._model = None
        for block in self.all_blocks():
            for compMap in block.components.components().itervalues():
                for comp in compMap.itervalues():
                    comp._parent = None

        # Monkey-patch for deepcopying weakrefs
        # Only required on Python <= 2.6
        import sys
        if sys.version_info[0] == 2 and sys.version_info[1] <= 6:
            copy._copy_dispatch[weakref.ref] = copy._copy_immutable
            copy._deepcopy_dispatch[weakref.ref] = copy._deepcopy_atomic
            copy._deepcopy_dispatch[weakref.KeyedRef] = copy._deepcopy_atomic

            def dcwvd(self, memo):
                """Deepcopy implementation for WeakValueDictionary class"""
                from copy import deepcopy
                new = self.__class__()
                for key, wr in self.data.items():
                    o = wr()
                    if o is not None:
                        new[deepcopy(key, memo)] = o
                return new
            weakref.WeakValueDictionary.__copy__ = weakref.WeakValueDictionary.copy
            weakref.WeakValueDictionary.__deepcopy__ = dcwvd

            def dcwkd(self, memo):
                """Deepcopy implementation for WeakKeyDictionary class"""
                from copy import deepcopy
                new = self.__class__()
                for key, value in self.data.items():
                    o = key()
                    if o is not none:
                        new[o] = deepcopy(value, memo)
                return new
            weakref.WeakKeyDictionary.__copy__ = weakref.WeakKeyDictionary.copy
            weakref.WeakKeyDictionary.__deepcopy__ = dcwkd

        # Actually do the copy
        instance = copy.deepcopy(self)
        #for name in self.components:
        #    self._model = weakref.ref(self)
        #    self.components[name]._model = weakref.ref(self)
        #for name in instance.components:
        #    instance._model = weakref.ref(instance)
        #    instance.components[name]._model = weakref.ref(instance)
        self._model = weakref.ref(self)
        instance._model = weakref.ref(instance)
        for block in self.all_blocks():
            for compMap in block.components.components().itervalues():
                for comp in compMap.itervalues():
                    comp._parent = weakref.ref(block)
        for block in instance.all_blocks():
            for compMap in block.components.components().itervalues():
                for comp in compMap.itervalues():
                    comp._parent = weakref.ref(block)
        return instance

    def reset(self):
        for name in self.components:
            self.components[name].reset()

    def preprocess(self):
        """Apply the preprocess plugins defined by the user"""
        suspend_gc = PauseGC()
        for item in self._preprocessors:
            self = Model.preprocessor_ep.service(item).preprocess(self)

    def solution(selfNone):
        """Return the solution"""
        soln = []
        for i in range(len(self._var)):
            soln.append( self._var[i].value )
        return soln

    # this method is a hack, used strictly by the pyomo command-line utility to
    # allow for user-readable names to be present in solver results objects.
    # this should be removed in the very near future, when labels are isolated
    # completely to solver plugins. in that situation, only human-readable
    # names will be present in a SolverResults object, and this method can
    # be removed.
    #
    # WEH - I think we need this method, even if the labeling method changes.
    #
    # The var/con/obj data is ordered by variable name, sorted by index.
    # This may differ from the declaration order, as well as the
    # instance order.  I haven't figured out how to efficiently generate
    # the declaration order.
    #
    # JDS - I think I agree with JP that this is no longer necessary
    #
    def update_results(self, results):

        results_symbol_map = results._symbol_map
        same_instance = results_symbol_map is not None and \
            results_symbol_map.instance is self

        new_results = SolverResults()
        new_results.problem = results.problem
        new_results.solver = results.solver

        for i in xrange(len(results.solution)):

            input_soln = results.solution(i+1)
            input_soln_variable = input_soln.variable
            input_soln_constraint = input_soln.constraint

            new_soln = Solution()
            new_soln.gap = input_soln.gap
            new_soln.status = input_soln.status
            #
            # Variables
            #
            vars = OrderedDict()
            tmp = {}
            for label, entry in input_soln_variable.iteritems():
                # NOTE: the following is a hack, to handle the ONE_VAR_CONSTANT variable
                if label == "ONE_VAR_CONSTANT":
                    continue
                # translate the label first if there is a symbol map
                # associated with the input solution.
                if same_instance:
                    var_value = results_symbol_map.getObject(label)
                elif results_symbol_map is not None:
                    var_value = results_symbol_map.getEquivalentObject(label, self)
                else:
                    raise RuntimeError("Cannot update from results missing a symbol map")

                if var_value is SymbolMap.UnknownSymbol:
                    msg = "Variable with label '%s' is not in model '%s'."
                    raise KeyError, msg % ( label, self.name, )

                if var_value.index.__class__ is tuple:
                    tmp[(var_value.component().name,)+var_value.index] = (var_value.name, entry)
                else:
                    tmp[(var_value.component().name,var_value.index)] = (var_value.name, entry)
                #vars[var_value.name] = entry
            for key in sorted(tmp.keys()):
                value = tmp[key]
                vars[value[0]] = value[1]
            new_soln.variable = vars
            #
            # Constraints
            #
            tmp = {}
            for label, entry in input_soln_constraint.iteritems():
                # NOTE: the following is a hack, to handle the ONE_VAR_CONSTANT variable
                if label == "c_e_ONE_VAR_CONSTANT":
                    continue
                if same_instance:
                    con_value = results_symbol_map.getObject(label)
                elif results_symbol_map is not None:
                    con_value = results_symbol_map.getEquivalentObject(label, self)
                else:
                    raise RuntimeError("Cannot update from results missing a symbol map")

                if con_value is SymbolMap.UnknownSymbol:
                    msg = "Constraint with label '%s' is not in model '%s'."
                    raise KeyError, msg % ( label, self.name, )

                if con_value.index.__class__ is tuple:
                    tmp[(con_value.component().name,)+con_value.index] = (con_value.name, entry)
                else:
                    tmp[(con_value.component().name,con_value.index)] = (con_value.name, entry)
            for key in sorted(tmp.keys()):
                value = tmp[key]
                new_soln.constraint.declare(value[0])
                dict.__setitem__(new_soln.constraint, value[0], value[1])
            #
            # Objectives
            #
            tmp = {}
            for label in input_soln.objective.keys():
                if same_instance:
                    obj_value = results_symbol_map.getObject(label)
                elif results_symbol_map is not None:
                    obj_value = results_symbol_map.getEquivalentObject(label, self)
                else:
                    raise RuntimeError("Cannot update from results missing a symbol map")

                if obj_value is SymbolMap.UnknownSymbol:
                    msg = "Objective with label '%s' is not in model '%s'."
                    raise KeyError, msg % ( label, self.name, )

                entry = input_soln.objective[label]
                if obj_value.index.__class__ is tuple:
                    tmp[(obj_value.component().name,)+obj_value.index] = (obj_value.name, entry)
                else:
                    tmp[(obj_value.component().name,obj_value.index)] = (obj_value.name, entry)
            for key in sorted(tmp.keys()):
                value = tmp[key]
                new_soln.objective.declare(value[0])
                dict.__setitem__(new_soln.objective, value[0], value[1])
            #
            new_results.solution.insert(new_soln)

        return new_results

    def load(self, arg, namespaces=[None], symbol_map=None,
             allow_consistent_values_for_fixed_vars=False,
             simplify=None, profile_memory=0, ignore_invalid_labels=False):
        """ Load the model with data from a file or a Solution object """

        if simplify is not None:
            print """
 WARNING: The 'simplify' argument to PyomoModel.load() has been deprecated.
      Please remove references from your code.
 """

        if arg is None or type(arg) is str:
            self._load_model_data(ModelData(filename=arg,model=self), namespaces, profile_memory=profile_memory)
            return True
        elif type(arg) is ModelData:
            self._load_model_data(arg, namespaces, profile_memory=profile_memory)
            return True
        elif type(arg) is coopr.opt.SolverResults:
            # if the solver status not one of either OK or Warning, then error.
            if (arg.solver.status != coopr.opt.SolverStatus.ok) and \
               (arg.solver.status != coopr.opt.SolverStatus.warning):

                if (arg.solver.status == coopr.opt.SolverStatus.aborted) and (len(arg.solution) > 0):
                   print "WARNING - Loading a SolverResults object with an 'aborted' status, but containing a solution"
                else:
                   msg = 'Cannot load a SolverResults object with bad status: %s'
                   raise ValueError, msg % str( arg.solver.status )

            # but if there is a warning, print out a warning, as someone should
            # probably take a look!
            if (arg.solver.status == coopr.opt.SolverStatus.warning):
                print 'WARNING - Loading a SolverResults object with a '       \
                      'warning status'

            if len(arg.solution) > 0:
                self._load_solution(
                    arg.solution(0),
                    symbol_map=arg.__dict__.get('_symbol_map', None),
                    allow_consistent_values_for_fixed_vars
                    = allow_consistent_values_for_fixed_vars,
                    ignore_invalid_labels=ignore_invalid_labels )
                return True
            else:
                return False
        elif type(arg) is coopr.opt.Solution:
            self._load_solution(
                arg,
                symbol_map=symbol_map,
                allow_consistent_values_for_fixed_vars
                = allow_consistent_values_for_fixed_vars,
                ignore_invalid_labels=ignore_invalid_labels )
            return True
        #elif type(arg) in (tuple, list):
        #    self._load_solution(
        #        arg,
        #        allow_consistent_values_for_fixed_vars
        #        = allow_consistent_values_for_fixed_vars,
        #        ignore_invalid_labels=ignore_invalid_labels )
        #    return True
        else:
            msg = "Cannot load model with object of type '%s'"
            raise ValueError, msg % str( type(arg) )

    def store_info(self, results):
        """ Store model information into a SolverResults object """
        results.problem.name = self.name

        stat_keys = (
          'number_of_variables',
          'number_of_binary_variables',
          'number_of_integer_variables',
          'number_of_continuous_variables',
          'number_of_constraints',
          'number_of_objectives'
        )

        for key in stat_keys:
            results.problem.__dict__[key] = self.statistics.__dict__[key]

    def _tuplize(self, data, setobj):
        if data is None:            #pragma:nocover
            return None
        if setobj.dimen == 1:
            return data
        ans = {}
        for key in data:
            if type(data[key][0]) is tuple:
                return data
            ans[key] = tuplize(data[key], setobj.dimen, setobj.name)
        return ans

    def _load_model_data(self, modeldata, namespaces, **kwds):
        """
        Load declarations from a ModelData object.
        """

        # As we are primarily generating objects here (and acyclic ones
        # at that), there is no need to run the GC until the entire
        # model is created.  Simple reference-counting should be
        # sufficient to keep memory use under control.
        suspend_gc = PauseGC()

        profile_memory = kwds.get('profile_memory', 0)

        # for pretty-printing summaries - unlike the standard
        # method in the pympler summary module, the tracker
        # doesn't print 0-byte entries to pad out the limit.
        if (pympler_available is True) and (profile_memory >= 2):
            memory_tracker = tracker.SummaryTracker()

        # Do some error checking
        for namespace in namespaces:
            if not namespace is None and not namespace in modeldata._data:
                msg = "Cannot access undefined namespace: '%s'"
                raise IOError, msg % namespace

        # Initialize each component in order.
        components = [key for key in self.components]

        for component_name in components:

            if (pympler_available is True) and (profile_memory >= 2):
                memory_tracker.create_summary()

            if self.components[component_name].type() is Model:
                continue
            elif self.components[component_name].type() is Var:
                var = self.components[component_name]
                if var._defer_domain:
                    domain_name = var.domain.name
                    var.domain = list(var.domain)
                    var.domain.sort()
                    if var.domain == []:
                        raise ValueError,"Attempting to set a variable's domain to an empty Set"
                    else:
                        if len(var.domain) != var.domain[-1] - var.domain[0] + 1:
                            raise ValueError,"Attempting to set a variable's domain to an improperly formatted Set " \
                                           + "-- elements are missing or duplicates exist: {0}".format(var.domain)
                        elif not all(type(elt) is int for elt in var.domain):
                            raise ValueError,"Attempting to set a veriable's domain to a Set with noninteger elements: {0}".format(var.domain)
                        else:
                            var.bounds = (var.domain[0],var.domain[-1])
                            var.domain = IntegerSet()
                            var.domain.name = domain_name
                            self.active_components(Var)[key] = var

            declaration = self.components[component_name]
            if component_name in modeldata._default.keys():
                if declaration.type() is Set:
                    declaration.set_default(self._tuplize(modeldata._default[component_name], declaration))
                else:
                    declaration.set_default(modeldata._default[component_name])
            data = None

            for namespace in namespaces:
                if component_name in modeldata._data.get(namespace,{}).keys():
                    if declaration.type() is Set:
                        data = self._tuplize(modeldata._data[namespace][component_name],
                                             declaration)
                    else:
                        data = modeldata._data[namespace][component_name]
                if not data is None:
                    break

            if __debug__:
                if logger.isEnabledFor(logging.DEBUG):
                    msg = "About to generate '%s' with data: %s"
                    print msg % ( declaration.name, str(data) )
                    self.pprint()

            try:
                declaration.construct(data)
            except:
                logger.error("constructing declaration '%s' from data=%s failed",
                             str(declaration.name), str(data).strip() )
                raise

            #print component_name, data # XXX

            if (pympler_available is True) and (profile_memory >= 2):
                print "Objects created during construction of component "+component_name+":"
                memory_tracker.print_diff()




    def _load_solution( self, soln, symbol_map,
                        allow_consistent_values_for_fixed_vars=False,
                        ignore_invalid_labels=False ):
        """
        Load a solution. A solution can be either a tuple or list, or a coopr.opt.Solution instance.
        - The allow_consistent_values_for_fixed_vars flag indicates whether a solution can specify
          consistent values for variables in the model that are fixed.
        - The ignore_invalid_labels flag indicates whether labels in the solution that don't
          appear in the model yield an error. This allows for loading a results object 
          generated from one model into another related, but not identical, model.
        """
        ## Load list or tuple into the variables, based on their order
        # NB: This is no longer supportable with the new SymbolMap 
        #if type(soln) in (tuple, list):
        #    if len(soln) != len(self._var):
        #        msg = 'Attempting to load a list/tuple solution into a ' \
        #              'model, but its length is %d while the model has %d ' \
        #              'variables' % (len(soln), len(self._var))
        #        raise ValueError, msg % ( len(soln), len(self._var) )
        #    for i in range(len(soln)):
        #        self._var[i].value = soln[i]
        #    return

        #
        # FIXME: JDS: Why doesn't this load the objective data?
        #

        if symbol_map is None:
            same_instance = False
        else:
            same_instance = symbol_map.instance is self
        #
        # Load variable data
        #
        for label, entry in soln.variable.iteritems():

            if same_instance:
                var_value = symbol_map.getObject(label)
            elif symbol_map is None:
                # We are going to assume the Solution was labeled with
                # the SolverResults pickler
                if 'canonical_label' in entry:
                    var_value = self.find_component(entry['canonical_label'])
                else:
                    # A last-ditch effort to resolve the object by the
                    # old lp-style labeling scheme
                    tmp = label
                    if tmp[-1] == ')': tmp = tmp[:-1]
                    tmp = tmp.replace('(',':').replace(')','.')
                    var_value = self.find_component(tmp)
            else:
                var_value = symbol_map.getEquivalentObject(label, self)

            if var_value is SymbolMap.UnknownSymbol:
                # NOTE: the following is a hack, to handle the ONE_VAR_CONSTANT
                #    variable that is necessary for the objective constant-offset
                #    terms.  probably should create a dummy variable in the model
                #    map at the same time the objective expression is being
                #    constructed.
                if label == "ONE_VAR_CONSTANT":
                    continue
                elif ignore_invalid_labels is True:
                    continue
                else:
                    raise KeyError("Variable label '%s' is not in model '%s'."
                                   % ( label, self.name, ))

            if not isinstance(var_value,_VarData):
                msg = "Variable '%s' in model '%s' is type %s"
                raise TypeError, msg % (
                    label, self.name, str(type(var_value)) )

            if (allow_consistent_values_for_fixed_vars is False) and (var_value.fixed is True):
                msg = "Variable '%s' in model '%s' is currently fixed - new" \
                      ' value is not expected in solution'
                raise TypeError, msg % ( label, self.name )

            #if var_value.status is not VarStatus.used:
            #    msg = "Variable '%s' in model '%s' is not currently used - no" \
            #          ' value is expected in solution'
            #    raise TypeError, msg % ( label, self.name )

            for attr_key, attr_value in entry.iteritems():
                if attr_key == 'Value':
                    if (allow_consistent_values_for_fixed_vars is True) and (var_value.fixed is True) and (attr_value != var_value.value):
                        msg = "Variable '%s' in model '%s' is currently fixed - a value of '%s' in solution is not identical to the current value of '%s'"
                        raise TypeError, msg % ( label, self.name, str(attr_value), str(var_value.value) )
                    var_value.value = attr_value
                elif attr_key != 'Id' and attr_key != 'canonical_label':
                    var_value.setattrvalue(attr_key, attr_value, define=True)
        #
        # Load constraint data
        #
        for label,entry in soln.constraint.iteritems():

            if same_instance:
                con_value = symbol_map.getObject(label)
            elif symbol_map is None:
                # We are going to assume the Solution was labeled with
                # the SolverResults pickler
                con_value = self.find_component(label)
            else:
                con_value = symbol_map.getEquivalentObject(label, self)

            if con_value is SymbolMap.UnknownSymbol:
                #
                # This is a hack - see above.
                #
                if label.endswith('ONE_VAR_CONSTANT'):
                    continue
                elif ignore_invalid_labels is True:
                    continue
                else:
                    raise KeyError("Constraint with label '%s' is not in model '%s'."
                                   % ( label, self.name, ))

            if not isinstance(con_value, _ConstraintData):
                raise TypeError("Constraint '%s' in model '%s' is type %s"
                                % ( label, self.name, str(type(con_value)) ))

            for _key in entry.iterkeys():
                key = _key[0].lower() + _key[1:]
                if key == 'value':
                    con_value.value = entry.value
                elif key != 'id' and key != 'canonical label':
                    setattr(con_value, key, getattr(entry, _key))

    def write(self, filename=None, format=ProblemFormat.cpxlp, solver_capability=None):
        """
        Write the model to a file, with a given format.
        """
        if format is None and not filename is None:
            #
            # Guess the format
            #
            format = guess_format(filename)
        #print "X",str(format),coopr.opt.WriterFactory.services()
        problem_writer = coopr.opt.WriterFactory(format)
        if problem_writer is None:
            msg = "Cannot write model in format '%s': no model writer "      \
                  "registered for that format"
            raise ValueError, msg % str(format)

        if solver_capability is None:
            solver_capability = lambda x: True
        (fname, symbol_map) = problem_writer(self, filename, solver_capability)
        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Writing model '%s' to file '%s' with format %s", self.name, str(fname), str(format))
        return fname, symbol_map

    def to_standard_form(self):
        """
        Produces a standard-form representation of the model. Returns
        the coefficient matrix (A), the cost vector (c), and the
        constraint vector (b), where the 'standard form' problem is

        min/max c'x
        s.t.    Ax = b
                x >= 0

        All three returned values are instances of the array.array
        class, and store Python floats (C doubles).
        """

        from coopr.pyomo.expr import generate_canonical_repn


        # We first need to create an map of all variables to their column
        # number
        colID = {}
        ID2name = {}
        id = 0
        tmp = self.variables().keys()
        tmp.sort()

        for v in tmp:
            colID[v] = id
            ID2name[id] = v
            id += 1

        # First we go through the constraints and introduce slack and excess
        # variables to eliminate inequality constraints
        #
        # N.B. Structure heirarchy:
        #
        # active_components: {class: {attr_name: object}}
        # object -> Constraint: ._data: {ndx: _ConstraintData}
        # _ConstraintData: .lower, .body, .upper
        #
        # So, altogether, we access a lower bound via
        #
        # model.active_components()[Constraint]['con_name']['index'].lower
        #
        # {le,ge,eq}Constraints are
        # {constraint_name: {index: {variable_or_none: coefficient}} objects
        # that represent each constraint. None in the innermost dictionary
        # represents the constant term.
        #
        # i.e.
        #
        # min  x1 + 2*x2 +          x4
        # s.t. x1                         = 1
        #           x2   + 3*x3          <= -1
        #      x1 +                 x4   >= 3
        #      x1 + 2*x2 +      +   3*x4 >= 0
        #
        #
        # would be represented as (modulo the names of the variables,
        # constraints, and indices)
        #
        # eqConstraints = {'c1': {None: {'x1':1, None:-1}}}
        # leConstraints = {'c2': {None: {'x2':1, 'x3':3, None:1}}}
        # geConstraints = {'c3': {None: {'x1':1, 'x4':1, None:-3}},
        #                  'c4': {None: {'x1':1, 'x2':2, 'x4':1, None:0}}}
        #
        # Note the we have the luxury of dealing only with linear terms.
        var_id_map = {}
        leConstraints = {}
        geConstraints = {}
        eqConstraints = {}
        objectives = {}
        # For each registered component
        for c in self.active_components():

            # Get all subclasses of Constraint
            if issubclass(c, Constraint):
                cons = self.active_components(c)

                # Get the name of the constraint, and the constraint set itself
                for con_set_name in cons:
                    con_set = cons[con_set_name]

                    # For each indexed constraint in the constraint set
                    for ndx in con_set._data:
                        con = con_set._data[ndx]

                        # Process the body
                        terms = self._process_canonical_repn(
                            generate_canonical_repn(con.body, var_id_map))

                        # Process the bounds of the constraint
                        if con._equality:
                            # Equality constraint, only check lower bound
                            lb = self._process_canonical_repn(
                                generate_canonical_repn(con.lower, var_id_map))

                            # Update terms
                            for k in lb:
                                v = lb[k]
                                if k in terms:
                                    terms[k] -= v
                                else:
                                    terms[k] = -v

                            # Add constraint to equality constraints
                            eqConstraints[(con_set_name, ndx)] = terms
                        else:

                            # Process upper bounds (<= constraints)
                            if con.upper is not None:
                                # Less than or equal to constraint
                                tmp = dict(terms)

                                ub = self._process_canonical_repn(
                                    generate_canonical_repn(con.upper, var_id_map))

                                # Update terms
                                for k in ub:
                                    if k in terms:
                                        tmp[k] -= ub[k]
                                    else:
                                        tmp[k] = -ub[k]

                                # Add constraint to less than or equal to
                                # constraints
                                leConstraints[(con_set_name, ndx)] = tmp

                            # Process lower bounds (>= constraints)
                            if con.lower is not None:
                                # Less than or equal to constraint
                                tmp = dict(terms)

                                lb = self._process_canonical_repn(
                                    generate_canonical_repn(con.lower, var_id_map))

                                # Update terms
                                for k in lb:
                                    if k in terms:
                                        tmp[k] -= lb[k]
                                    else:
                                        tmp[k] = -lb[k]

                                # Add constraint to less than or equal to
                                # constraints
                                geConstraints[(con_set_name, ndx)] = tmp
            elif issubclass(c, Objective):
                # Process objectives
                objs = self.active_components(c)

                # Get the name of the objective, and the objective set itself
                for obj_set_name in objs:
                    obj_set = objs[obj_set_name]

                    # For each indexed objective in the objective set
                    for ndx in obj_set._data:
                        obj = obj_set._data[ndx]
                        # Process the objective
                        terms = self._process_canonical_repn(
                            generate_canonical_repn(obj.expr, var_id_map))

                        objectives[(obj_set_name, ndx)] = terms


        # We now have all the constraints. Add a slack variable for every
        # <= constraint and an excess variable for every >= constraint.
        nSlack = len(leConstraints)
        nExcess = len(geConstraints)

        nConstraints = len(leConstraints) + len(geConstraints) + \
                       len(eqConstraints)
        nVariables = len(colID) + nSlack + nExcess
        nRegVariables = len(colID)

        # Make the arrays
        coefficients = array.array("d", [0]*nConstraints*nVariables)
        constraints = array.array("d", [0]*nConstraints)
        costs = array.array("d", [0]*nVariables)

        # Populate the coefficient matrix
        constraintID = 0

        # Add less than or equal to constraints
        for ndx in leConstraints:
            con = leConstraints[ndx]
            for termKey in con:
                coef = con[termKey]

                if termKey is None:
                    # Constraint coefficient
                    constraints[constraintID] = -coef
                else:
                    # Variable coefficient
                    col = colID[termKey]
                    coefficients[constraintID*nVariables + col] = coef

            # Add the slack
            coefficients[constraintID*nVariables + nRegVariables + \
                        constraintID] = 1
            constraintID += 1

        # Add greater than or equal to constraints
        for ndx in geConstraints:
            con = geConstraints[ndx]
            for termKey in con:
                coef = con[termKey]

                if termKey is None:
                    # Constraint coefficient
                    constraints[constraintID] = -coef
                else:
                    # Variable coefficient
                    col = colID[termKey]
                    coefficients[constraintID*nVariables + col] = coef

            # Add the slack
            coefficients[constraintID*nVariables + nRegVariables + \
                        constraintID] = -1
            constraintID += 1

        # Add equality constraints
        for ndx in eqConstraints:
            con = eqConstraints[ndx]
            for termKey in con:
                coef = con[termKey]

                if termKey is None:
                    # Constraint coefficient
                    constraints[constraintID] = -coef
                else:
                    # Variable coefficient
                    col = colID[termKey]
                    coefficients[constraintID*nVariables + col] = coef

            constraintID += 1

        # Determine cost coefficients
        for obj_name in objectives:
            obj = objectives[obj_name]()
            for var in obj:
                costs[colID[var]] = obj[var]

        # Print the model
        #
        # The goal is to print
        #
        #         var1   var2   var3   ...
        #       +--                     --+
        #       | cost1  cost2  cost3  ...|
        #       +--                     --+
        #       +--                     --+ +-- --+
        # con1  | coef11 coef12 coef13 ...| | eq1 |
        # con2  | coef21 coef22 coef23 ...| | eq2 |
        # con2  | coef31 coef32 coef33 ...| | eq3 |
        #  .    |   .      .      .   .   | |  .  |
        #  .    |   .      .      .    .  | |  .  |
        #  .    |   .      .      .     . | |  .  |

        constraintPadding = 2
        numFmt = "% 1.4f"
        altFmt = "% 1.1g"
        maxColWidth = max(len(numFmt % 0.0), len(altFmt % 0.0))
        maxConstraintColWidth = max(len(numFmt % 0.0), len(altFmt % 0.0))

        # Generate constraint names
        maxConNameLen = 0
        conNames = []
        for name in leConstraints:
            strName = str(name)
            if len(strName) > maxConNameLen:
                maxConNameLen = len(strName)
            conNames.append(strName)
        for name in geConstraints:
            strName = str(name)
            if len(strName) > maxConNameLen:
                maxConNameLen = len(strName)
            conNames.append(strName)
        for name in eqConstraints:
            strName = str(name)
            if len(strName) > maxConNameLen:
                maxConNameLen = len(strName)
            conNames.append(strName)

        # Generate the variable names
        varNames = [None]*len(colID)
        for name in colID:
            tmp_name = " " + name
            if len(tmp_name) > maxColWidth:
                maxColWidth = len(tmp_name)
            varNames[colID[name]] = tmp_name
        for i in xrange(0, nSlack):
            tmp_name = " _slack_%i" % i
            if len(tmp_name) > maxColWidth:
                maxColWidth = len(tmp_name)
            varNames.append(tmp_name)
        for i in xrange(0, nExcess):
            tmp_name = " _excess_%i" % i
            if len(tmp_name) > maxColWidth:
                maxColWidth = len(tmp_name)
            varNames.append(tmp_name)

        # Variable names
        line = " "*maxConNameLen + (" "*constraintPadding) + " "
        for col in xrange(0, nVariables):
            # Format entry
            token = varNames[col]

            # Pad with trailing whitespace
            token += " "*(maxColWidth - len(token))

            # Add to line
            line += " " + token + " "
        print line

        # Cost vector
        print " "*maxConNameLen + (" "*constraintPadding) + "+--" + \
              " "*((maxColWidth+2)*nVariables - 4) + "--+"
        line = " "*maxConNameLen + (" "*constraintPadding) + "|"
        for col in xrange(0, nVariables):
            # Format entry
            token = numFmt % costs[col]
            if len(token) > maxColWidth:
                token = altFmt % costs[col]

            # Pad with trailing whitespace
            token += " "*(maxColWidth - len(token))

            # Add to line
            line += " " + token + " "
        line += "|"
        print line
        print " "*maxConNameLen + (" "*constraintPadding) + "+--" + \
              " "*((maxColWidth+2)*nVariables - 4) + "--+"

        # Constraints
        print " "*maxConNameLen + (" "*constraintPadding) + "+--" + \
              " "*((maxColWidth+2)*nVariables - 4) + "--+" + \
              (" "*constraintPadding) + "+--" + \
              (" "*(maxConstraintColWidth-1)) + "--+"
        for row in xrange(0, nConstraints):
            # Print constraint name
            line = conNames[row] + (" "*constraintPadding) + (" "*(maxConNameLen - len(conNames[row]))) + "|"

            # Print each coefficient
            for col in xrange(0, nVariables):
                # Format entry
                token = numFmt % coefficients[nVariables*row + col]
                if len(token) > maxColWidth:
                    token = altFmt % coefficients[nVariables*row + col]

                # Pad with trailing whitespace
                token += " "*(maxColWidth - len(token))

                # Add to line
                line += " " + token + " "

            line += "|" + (" "*constraintPadding) + "|"

            # Add constraint vector
            token = numFmt % constraints[row]
            if len(token) > maxConstraintColWidth:
                token = altFmt % constraints[row]

            # Pad with trailing whitespace
            token += " "*(maxConstraintColWidth - len(token))

            line += " " + token + "  |"
            print line
        print " "*maxConNameLen + (" "*constraintPadding) + "+--" + \
              " "*((maxColWidth+2)*nVariables - 4) + "--+" + \
              (" "*constraintPadding) + "+--" + (" "*(maxConstraintColWidth-1))\
              + "--+"

        return (coefficients, costs, constraints)

    def _process_canonical_repn(self, expr):
        """
        Returns a dictionary of {var_name_or_None: coef} values
        """

        terms = {}

        # Get the variables from the canonical representation
        vars = expr.pop(-1, {})

        # Find the linear terms
        linear = expr.pop(1, {})
        for k in linear:
            # FrozeDicts don't support (k, v)-style iteration
            v = linear[k]

            # There's exactly 1 variable in each term
            terms[vars[k.keys()[0]].label] = v

        # Get the constant term, if present
        const = expr.pop(0, {})
        if None in const:
            terms[None] = const[None]

        if len(expr) != 0:
            raise TypeError, "Nonlinear terms in expression"

        return terms


def _tempTrue(x):
    """
    Defined to return True to all inquiries
    """
    return True


class ConcreteModel(Model):
    """
    A concrete optimization model that does not defer construction of
    components.
    """

    def __init__(self, *args, **kwds):
        kwds['_deprecate'] = False
        Model.__init__(self, *args, **kwds)
        self._defer_construction=False

    pyutilib.component.core.alias("ConcreteModel", 'A concrete optimization model that '\
              'does not defer construction of components.')

class AbstractModel(Model):
    """
    An abstract optimization model that defers construction of
    components.
    """

    def __init__(self, *args, **kwds):
        kwds['_deprecate'] = False
        Model.__init__(self, *args, **kwds)
        self._defer_construction=True

    pyutilib.component.core.alias("AbstractModel", 'An abstract optimization model that '\
              'defers construction of components.')
